# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import re

from drgn import Object, offsetof
from drgn.commands import CommandArgumentError
from drgn.commands._crash._collections import _find_tree_type
from drgn.helpers.linux.rbtree import (
    rbtree_inorder_for_each,
    rbtree_inorder_for_each_entry,
    rbtree_preorder_for_each,
    rbtree_preorder_for_each_entry,
)
from tests import TestCase
from tests.linux_kernel import skip_unless_have_test_kmod
from tests.linux_kernel.crash_commands import CrashCommandTestCase


@skip_unless_have_test_kmod
class TestList(CrashCommandTestCase):
    def test_null_terminated(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_custom_list_entry.next drgn_test_custom_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_custom_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in self.prog["drgn_test_custom_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn('prog["drgn_test_custom_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry.next", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )

    def test_end(self):
        end = self.prog["drgn_test_custom_list_entries"][1].address_
        cmd = self.check_crash_command(
            f"list -o drgn_test_custom_list_entry.next -e {end:#x} drgn_test_custom_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_custom_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in self.prog["drgn_test_custom_list_entries"][:1]
                    ),
                ]
            ),
        )

        self.assertIn('prog["drgn_test_custom_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry.next", cmd.drgn_option.stdout)
        self.assertIn(f"= {end:#x}", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )

    def test_no_offset(self):
        cmd = self.check_crash_command("list drgn_test_full_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_full_list'].address_:x}\n",
                    *(
                        f"{entry.node.address_:x}\n"
                        for entry in self.prog["drgn_test_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_full_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(ptr)", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["ptr"], int)

    def test_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_custom_list_entry"), "next")
        cmd = self.check_crash_command(f"list -o {offset} drgn_test_custom_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_custom_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in self.prog["drgn_test_custom_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_custom_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(ptr + next_offset)", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["ptr"], int)

    def test_circular(self):
        address = self.prog["drgn_test_custom_list_cycle"][2].address_
        cmd = self.check_crash_command(
            f"list -o drgn_test_custom_list_entry.next {address:#x}"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_custom_list_cycle"][2:]
                ]
            ),
        )

        self.assertIn(
            f'Object(prog, "struct drgn_test_custom_list_entry *", {address:#x})',
            cmd.drgn_option.stdout,
        )
        self.assertIn("entry.next", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )

    def test_circular_next(self):
        address = self.prog["drgn_test_custom_list_cycle"][1].address_
        cmd = self.run_crash_command(
            f"list -o drgn_test_custom_list_entry.next {address:#x}"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_custom_list_cycle"][1:]
                ]
            ),
        )

    def test_duplicate(self):
        cmd = self.run_crash_command(
            "list -o drgn_test_custom_list_entry.next drgn_test_custom_list_cycle"
        )
        expected_lines = [
            f"{entry.address_:x}\n"
            for entry in self.prog["drgn_test_custom_list_cycle"]
        ]
        expected_lines.append(expected_lines[2])
        expected_lines.append("\nlist: duplicate list entry: " + expected_lines[-1])
        self.assertEqual(cmd.stdout, "".join(expected_lines))

    def test_self_cycle(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_custom_list_entry.next drgn_test_custom_list_self_cycle"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_custom_list_self_cycle"]
                ]
            ),
        )

        # drgn_test_custom_list_self_cycle is an array, not a struct
        # drgn_test_custom_list_entry, so --drgn falls back to a symbol lookup.
        self.assertIn(
            'prog.symbol("drgn_test_custom_list_self_cycle").address',
            cmd.drgn_option.stdout,
        )
        self.assertIn("entry.next", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )

    def test_struct(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_custom_list_entry.next -s drgn_test_custom_list_entry drgn_test_custom_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_custom_list"],
                *self.prog["drgn_test_custom_list_entries"],
            ],
            1,
        ):
            regex.append(
                rf"{entry.address_:x}\n\(struct drgn_test_custom_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_custom_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry.next", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )

    def test_struct_member(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_custom_list_entry.next -s drgn_test_custom_list_entry.value drgn_test_custom_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_custom_list"],
                *self.prog["drgn_test_custom_list_entries"],
            ],
            1,
        ):
            regex.append(rf"{entry.address_:x}\n\s*value = \(int\){i}\n")
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_custom_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry.next", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )
        self.assertIsInstance(cmd.drgn_option.globals["value"], Object)

    def test_struct_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_custom_list_entry"), "next")
        cmd = self.check_crash_command(
            f"list -o {offset} -s drgn_test_custom_list_entry drgn_test_custom_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_custom_list"],
                *self.prog["drgn_test_custom_list_entries"],
            ],
            1,
        ):
            regex.append(
                rf"{entry.address_:x}\n\(struct drgn_test_custom_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn(
            'prog.symbol("drgn_test_custom_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(ptr + next_offset)", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )

    def test_struct_member_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_custom_list_entry"), "next")
        cmd = self.check_crash_command(
            f"list -o {offset} -s drgn_test_custom_list_entry.value drgn_test_custom_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_custom_list"],
                *self.prog["drgn_test_custom_list_entries"],
            ],
            1,
        ):
            regex.append(rf"{entry.address_:x}\n\s*value = \(int\){i}\n")
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn(
            'prog.symbol("drgn_test_custom_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(ptr + next_offset)", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_custom_list_entry *",
        )
        self.assertIsInstance(cmd.drgn_option.globals["value"], Object)

    def test_list_head_empty(self):
        cmd = self.run_crash_command("list -H drgn_test_empty_list")
        self.assertEqual(cmd.stdout, "(empty)\n")

    def test_list_head_full(self):
        cmd = self.check_crash_command("list -H drgn_test_full_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_list_entries"]
                ]
            ),
        )

        self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
        self.assertIn("list_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_list_head_address(self):
        address = self.prog["drgn_test_full_list"].address_
        cmd = self.check_crash_command(f"list -H {address:#x}")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_list_entries"]
                ]
            ),
        )

        self.assertIn(
            f'Object(prog, "struct list_head *", {address:#x})', cmd.drgn_option.stdout
        )
        self.assertIn("list_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_list_head_reverse(self):
        cmd = self.check_crash_command("list -r -H drgn_test_full_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in reversed(self.prog["drgn_test_list_entries"])
                ]
            ),
        )

        self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_reverse(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_list_head_cycle1(self):
        cmd = self.run_crash_command("list -H drgn_test_list_cycle1")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_list_cycle1_entries"]
                ]
            ),
        )

    def test_list_head_cycle2(self):
        cmd = self.run_crash_command("list -H drgn_test_list_cycle2")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_list_cycle2_entries"]
                ]
            ),
        )

    def test_list_head_cycle3(self):
        cmd = self.run_crash_command("list -H drgn_test_list_cycle3")
        expected_lines = [
            f"{entry.node.address_:x}\n"
            for entry in self.prog["drgn_test_list_cycle3_entries"]
        ]
        expected_lines.append(expected_lines[2])
        expected_lines.append("\nlist: duplicate list entry: " + expected_lines[-1])
        self.assertEqual(cmd.stdout, "".join(expected_lines))

    def test_list_head_o(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -H drgn_test_full_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_list_entries"]
                ]
            ),
        )

        self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_list_head_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(f"list -o {offset} -H drgn_test_full_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_list_entries"]
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_full_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(start)", cmd.drgn_option.stdout)
        self.assertIn("entry_ptr = ptr - list_head_offset\n", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["list_head_offset"], offset)
        self.assertIn("while ptr != start:", cmd.drgn_option.stdout)
        self.assertNotIn("not next_ptr", cmd.drgn_option.stdout)

    def test_list_head_reverse_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(f"list -o {offset} -r -H drgn_test_full_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in reversed(self.prog["drgn_test_list_entries"])
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_full_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(start + prev_offset)", cmd.drgn_option.stdout)
        self.assertIn("while ptr != start:", cmd.drgn_option.stdout)
        self.assertNotIn("not prev_ptr", cmd.drgn_option.stdout)

    def test_list_head_struct(self):
        cmd = self.check_crash_command(
            "list -s drgn_test_list_entry -l drgn_test_list_entry.node -H drgn_test_full_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(self.prog["drgn_test_list_entries"], 1):
            regex.append(
                rf"{entry.node.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_list_head_struct_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"list -s drgn_test_list_entry -l {offset} -H drgn_test_full_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(self.prog["drgn_test_list_entries"], 1):
            regex.append(
                rf"{entry.node.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn(
            'prog.symbol("drgn_test_full_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(start)", cmd.drgn_option.stdout)
        self.assertIn(
            'entry = Object(prog, "struct drgn_test_list_entry *", ptr - entry_offset)',
            cmd.drgn_option.stdout,
        )
        self.assertIn("while ptr != start:", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["entry_offset"], offset)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_list_head_struct_o(self):
        cmd = self.check_crash_command(
            "list -s drgn_test_list_entry -o drgn_test_list_entry.node -H drgn_test_full_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(self.prog["drgn_test_list_entries"], 1):
            regex.append(
                rf"{entry.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_list_head_struct_o_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"list -s drgn_test_list_entry -o {offset} -H drgn_test_full_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(self.prog["drgn_test_list_entries"], 1):
            regex.append(
                rf"{entry.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn(
            'prog.symbol("drgn_test_full_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(start)", cmd.drgn_option.stdout)
        self.assertIn(
            'entry = Object(prog, "struct drgn_test_list_entry *", ptr - list_head_offset)',
            cmd.drgn_option.stdout,
        )
        self.assertIn("while ptr != start:", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["list_head_offset"], offset)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_list_head_struct_reverse(self):
        cmd = self.check_crash_command(
            "list -s drgn_test_list_entry -l drgn_test_list_entry.node -r -H drgn_test_full_list"
        )
        regex = [r"^"]
        for i, entry in reversed(
            list(enumerate(self.prog["drgn_test_list_entries"], 1))
        ):
            regex.append(
                rf"{entry.node.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry_reverse(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_list_head_struct_reverse_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"list -s drgn_test_list_entry -l {offset} -r -H drgn_test_full_list"
        )
        regex = [r"^"]
        for i, entry in reversed(
            list(enumerate(self.prog["drgn_test_list_entries"], 1))
        ):
            regex.append(
                rf"{entry.node.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn(
            'prog.symbol("drgn_test_full_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("prog.read_word(start + prev_offset)", cmd.drgn_option.stdout)
        self.assertIn(
            'entry = Object(prog, "struct drgn_test_list_entry *", ptr - entry_offset)',
            cmd.drgn_option.stdout,
        )
        self.assertIn("while ptr != start:", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["entry_offset"], offset)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_embedded_list_head(self):
        cmd = self.check_crash_command("list -h drgn_test_full_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_full_list'].address_:x}\n",
                    *(
                        f"{entry.node.address_:x}\n"
                        for entry in self.prog["drgn_test_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
        self.assertIn("node = node.next", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_embedded_list_head_address(self):
        address = self.prog["drgn_test_full_list"].address_
        cmd = self.check_crash_command(f"list -h {address:#x}")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_full_list'].address_:x}\n",
                    *(
                        f"{entry.node.address_:x}\n"
                        for entry in self.prog["drgn_test_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn(
            f'Object(prog, "struct list_head *", {address:#x})', cmd.drgn_option.stdout
        )
        self.assertIn("node = node.next", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_embedded_list_head_o(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -h drgn_test_circular_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_circular_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in self.prog["drgn_test_circular_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn('prog["drgn_test_circular_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry = list_next_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_embedded_list_head_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(f"list -o {offset} -h drgn_test_circular_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_circular_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in self.prog["drgn_test_circular_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_circular_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("entry_ptr = ptr - list_head_offset\n", cmd.drgn_option.stdout)
        self.assertIn("prog.read_word(ptr)", cmd.drgn_option.stdout)
        self.assertNotIn("not next_ptr", cmd.drgn_option.stdout)

    def test_embedded_list_head_address_o(self):
        address = self.prog["drgn_test_circular_list"].address_
        cmd = self.check_crash_command(
            f"list -o drgn_test_list_entry.node -h {address:#x}"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_circular_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in self.prog["drgn_test_circular_list_entries"]
                    ),
                ]
            ),
        )

        self.assertIn(
            f'Object(prog, "struct drgn_test_list_entry *", {address:#x})',
            cmd.drgn_option.stdout,
        )
        self.assertIn("entry = list_next_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_embedded_list_head_reverse(self):
        for h in ("-h ", ""):
            with self.subTest("explicit" if h else "implicit"):
                cmd = self.check_crash_command(f"list -r {h}drgn_test_full_list")
                self.assertEqual(
                    cmd.stdout,
                    "".join(
                        [
                            f"{self.prog['drgn_test_full_list'].address_:x}\n",
                            *(
                                f"{entry.node.address_:x}\n"
                                for entry in reversed(
                                    self.prog["drgn_test_list_entries"]
                                )
                            ),
                        ]
                    ),
                )

                self.assertIn('prog["drgn_test_full_list"]', cmd.drgn_option.stdout)
                self.assertIn("node = node.prev", cmd.drgn_option.stdout)
                self.assertEqual(
                    cmd.drgn_option.globals["node"].type_.type_name(),
                    "struct list_head *",
                )

    def test_embedded_list_head_reverse_o(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -r -h drgn_test_circular_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_circular_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in reversed(
                            self.prog["drgn_test_circular_list_entries"]
                        )
                    ),
                ]
            ),
        )

        self.assertIn('prog["drgn_test_circular_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry = list_prev_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_embedded_list_head_reverse_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"list -o {offset} -r -h drgn_test_circular_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{self.prog['drgn_test_circular_list'].address_:x}\n",
                    *(
                        f"{entry.address_:x}\n"
                        for entry in reversed(
                            self.prog["drgn_test_circular_list_entries"]
                        )
                    ),
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_circular_list").address', cmd.drgn_option.stdout
        )
        self.assertIn("entry_ptr = ptr - list_head_offset\n", cmd.drgn_option.stdout)
        self.assertIn("prog.read_word(ptr + prev_offset)", cmd.drgn_option.stdout)
        self.assertNotIn("not prev_ptr", cmd.drgn_option.stdout)

    def test_embedded_list_head_struct(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -s drgn_test_list_entry -h drgn_test_circular_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_circular_list"],
                *self.prog["drgn_test_circular_list_entries"],
            ],
            1,
        ):
            regex.append(
                rf"{entry.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_circular_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry = list_next_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_embedded_list_head_member(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -s drgn_test_list_entry.value -h drgn_test_circular_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_circular_list"],
                *self.prog["drgn_test_circular_list_entries"],
            ],
            1,
        ):
            regex.append(rf"{entry.address_:x}\n\s*value = \(int\){i}\n")
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_circular_list"]', cmd.drgn_option.stdout)
        self.assertIn("entry = list_next_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )
        self.assertIsInstance(cmd.drgn_option.globals["value"], Object)

    def test_embedded_list_head_struct_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"list -o {offset} -s drgn_test_list_entry -h drgn_test_circular_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_circular_list"],
                *self.prog["drgn_test_circular_list_entries"],
            ],
            1,
        ):
            regex.append(
                rf"{entry.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn(
            'prog.symbol("drgn_test_circular_list").address', cmd.drgn_option.stdout
        )
        self.assertIn(
            'entry = Object(prog, "struct drgn_test_list_entry *", ptr - list_head_offset)',
            cmd.drgn_option.stdout,
        )
        self.assertEqual(cmd.drgn_option.globals["list_head_offset"], offset)
        self.assertIn("prog.read_word(ptr)", cmd.drgn_option.stdout)
        self.assertNotIn("not next_ptr", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_embedded_list_head_member_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"list -o {offset} -s drgn_test_list_entry.value -h drgn_test_circular_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(
            [
                self.prog["drgn_test_circular_list"],
                *self.prog["drgn_test_circular_list_entries"],
            ],
            1,
        ):
            regex.append(rf"{entry.address_:x}\n\s*value = \(int\){i}\n")
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn(
            'prog.symbol("drgn_test_circular_list").address', cmd.drgn_option.stdout
        )
        self.assertIn(
            'entry = Object(prog, "struct drgn_test_list_entry *", ptr - list_head_offset)',
            cmd.drgn_option.stdout,
        )
        self.assertEqual(cmd.drgn_option.globals["list_head_offset"], offset)
        self.assertIn("prog.read_word(ptr)", cmd.drgn_option.stdout)
        self.assertNotIn("not next_ptr", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )
        self.assertIsInstance(cmd.drgn_option.globals["value"], Object)

    def test_head_node_offset(self):
        cmd = self.check_crash_command(
            "list -O drgn_test_list_anchor.list -h drgn_test_anchored_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_anchored_list_entries"]
                ]
            ),
        )

        self.assertIn('prog["drgn_test_anchored_list"].list', cmd.drgn_option.stdout)
        self.assertIn("list_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_head_node_integer_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_anchor"), "list")
        cmd = self.check_crash_command(f"list -O {offset} -h drgn_test_anchored_list")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_anchored_list_entries"]
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_anchored_list").address', cmd.drgn_option.stdout
        )
        self.assertEqual(cmd.drgn_option.globals["offset"], offset)
        self.assertIn(
            'Object(prog, "struct list_head *", address + offset)',
            cmd.drgn_option.stdout,
        )
        self.assertIn("list_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_head_node_offset_address(self):
        address = self.prog["drgn_test_anchored_list"].address_
        cmd = self.check_crash_command(
            f"list -O drgn_test_list_anchor.list -h {address:#x}"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_anchored_list_entries"]
                ]
            ),
        )

        self.assertIn(
            f'Object(prog, "struct drgn_test_list_anchor", address={address:#x}).list',
            cmd.drgn_option.stdout,
        )
        self.assertIn("list_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_head_node_integer_offset_address(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_anchor"), "list")
        address = self.prog["drgn_test_anchored_list"].address_
        cmd = self.check_crash_command(f"list -O {offset} -h {address:#x}")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in self.prog["drgn_test_anchored_list_entries"]
                ]
            ),
        )

        self.assertEqual(cmd.drgn_option.globals["offset"], offset)
        self.assertIn(
            f'Object(prog, "struct list_head *", {address:#x} + offset)',
            cmd.drgn_option.stdout,
        )
        self.assertIn("list_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_head_node_offset_o(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -O drgn_test_list_anchor.list -h drgn_test_anchored_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_anchored_list_entries"]
                ]
            ),
        )

        self.assertIn('prog["drgn_test_anchored_list"].list', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_head_node_offset_and_integer_o(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        cmd = self.check_crash_command(
            f"list -o {offset} -O drgn_test_list_anchor.list -h drgn_test_anchored_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_anchored_list_entries"]
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_anchored_list").address', cmd.drgn_option.stdout
        )
        self.assertIn(
            'start += offsetof("struct drgn_test_list_anchor", "list")',
            cmd.drgn_option.stdout,
        )
        self.assertIn("prog.read_word(start)", cmd.drgn_option.stdout)
        self.assertIn("entry_ptr = ptr - list_head_offset\n", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["list_head_offset"], offset)
        self.assertNotIn("not next_ptr", cmd.drgn_option.stdout)

    def test_head_node_integer_offset_and_integer_o(self):
        offset = offsetof(self.prog.type("struct drgn_test_list_entry"), "node")
        head_node_offset = offsetof(
            self.prog.type("struct drgn_test_list_anchor"), "list"
        )
        cmd = self.check_crash_command(
            f"list -o {offset} -O {head_node_offset} -h drgn_test_anchored_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in self.prog["drgn_test_anchored_list_entries"]
                ]
            ),
        )

        self.assertIn(
            'prog.symbol("drgn_test_anchored_list").address', cmd.drgn_option.stdout
        )
        self.assertIn(f"start += {head_node_offset}", cmd.drgn_option.stdout)
        self.assertIn("prog.read_word(start)", cmd.drgn_option.stdout)
        self.assertIn("entry_ptr = ptr - list_head_offset\n", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["list_head_offset"], offset)
        self.assertNotIn("not next_ptr", cmd.drgn_option.stdout)

    def test_head_node_offset_struct(self):
        cmd = self.check_crash_command(
            "list -l drgn_test_list_entry.node -s drgn_test_list_entry -O drgn_test_list_anchor.list -h drgn_test_anchored_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(self.prog["drgn_test_anchored_list_entries"], 1):
            regex.append(
                rf"{entry.node.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_anchored_list"].list', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_head_node_offset_struct_o(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -s drgn_test_list_entry -O drgn_test_list_anchor.list -h drgn_test_anchored_list"
        )
        regex = [r"^"]
        for i, entry in enumerate(self.prog["drgn_test_anchored_list_entries"], 1):
            regex.append(
                rf"{entry.address_:x}\n\(struct drgn_test_list_entry\)(?s:.)*\.value = \(int\){i},(?s:.)*"
            )
        regex.append("$")
        self.assertRegex(cmd.stdout, "".join(regex))

        self.assertIn('prog["drgn_test_anchored_list"].list', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )

    def test_head_node_offset_reverse(self):
        cmd = self.check_crash_command(
            "list -O drgn_test_list_anchor.list -r -h drgn_test_anchored_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.node.address_:x}\n"
                    for entry in reversed(self.prog["drgn_test_anchored_list_entries"])
                ]
            ),
        )

        self.assertIn('prog["drgn_test_anchored_list"].list', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_reverse(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct list_head *"
        )

    def test_head_node_offset_reverse_o(self):
        cmd = self.check_crash_command(
            "list -o drgn_test_list_entry.node -O drgn_test_list_anchor.list -r -h drgn_test_anchored_list"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.address_:x}\n"
                    for entry in reversed(self.prog["drgn_test_anchored_list_entries"])
                ]
            ),
        )

        self.assertIn('prog["drgn_test_anchored_list"].list', cmd.drgn_option.stdout)
        self.assertIn("list_for_each_entry_reverse(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_list_entry *",
        )


class TestFindTreeType(TestCase):
    def test_full(self):
        self.assertEqual(_find_tree_type("rbtree"), "rbtree")
        self.assertEqual(_find_tree_type("radix"), "radix")
        self.assertEqual(_find_tree_type("xarray"), "xarray")
        self.assertEqual(_find_tree_type("maple"), "maple")

    def test_abbreviated(self):
        for expected, abbreviations in (
            ("rbtree", ("rb", "rbt", "rbtr", "rbtre")),
            ("radix", ("ra", "rad", "radi")),
            ("xarray", ("x", "xa", "xar", "xarr", "xarra")),
            ("maple", ("m", "ma", "map", "mapl")),
        ):
            for abbreviation in abbreviations:
                with self.subTest(abbreviation=abbreviation):
                    self.assertEqual(_find_tree_type(abbreviation), expected)

    def test_ambiguous(self):
        self.assertRaises(CommandArgumentError, _find_tree_type, "r")

    def test_unknown(self):
        self.assertRaises(CommandArgumentError, _find_tree_type, "z")


@skip_unless_have_test_kmod
class TestTree(CrashCommandTestCase):
    def test_rbtree(self):
        cmd = self.check_crash_command("tree drgn_test_rb_root")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{node.value_():x}\n"
                    for node in rbtree_preorder_for_each(
                        self.prog["drgn_test_rb_root"].address_of_()
                    )
                ]
            ),
        )
        self.assertIn("root = prog[", cmd.drgn_option.stdout)
        self.assertIn("rbtree_preorder_for_each(", cmd.drgn_option.stdout)
        self.assertNotIn("rbtree_inorder_for_each", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_rb_root"].address_of_(),
        )
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct rb_node *"
        )

        with self.subTest("explicit type"):
            explicit_cmd = self.check_crash_command("tree -t rbtree drgn_test_rb_root")
            self.assertEqual(explicit_cmd.stdout, cmd.stdout)
            self.assertEqual(explicit_cmd.drgn_option.stdout, cmd.drgn_option.stdout)

    def test_rbtree_linear(self):
        cmd = self.check_crash_command("tree -l drgn_test_rb_root")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{node.value_():x}\n"
                    for node in rbtree_inorder_for_each(
                        self.prog["drgn_test_rb_root"].address_of_()
                    )
                ]
            ),
        )
        self.assertIn("root = prog[", cmd.drgn_option.stdout)
        self.assertIn("rbtree_inorder_for_each(", cmd.drgn_option.stdout)
        self.assertNotIn("rbtree_preorder_for_each", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_rb_root"].address_of_(),
        )
        self.assertEqual(
            cmd.drgn_option.globals["node"].type_.type_name(), "struct rb_node *"
        )

    def test_rbtree_node_member(self):
        cmd = self.check_crash_command(
            "tree -o drgn_test_rb_entry.node drgn_test_rb_root"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.value_():x}\n"
                    for entry in rbtree_preorder_for_each_entry(
                        "struct drgn_test_rb_entry",
                        self.prog["drgn_test_rb_root"].address_of_(),
                        "node",
                    )
                ]
            ),
        )
        self.assertIn("rbtree_preorder_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_rb_entry *",
        )

    def test_rbtree_node_member_linear(self):
        cmd = self.check_crash_command(
            "tree -l -o drgn_test_rb_entry.node drgn_test_rb_root"
        )
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.value_():x}\n"
                    for entry in rbtree_inorder_for_each_entry(
                        "struct drgn_test_rb_entry",
                        self.prog["drgn_test_rb_root"].address_of_(),
                        "node",
                    )
                ]
            ),
        )
        self.assertIn("rbtree_inorder_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_rb_entry *",
        )

    def test_rbtree_node_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_rb_entry"), "node")
        cmd = self.check_crash_command(f"tree -o {offset} drgn_test_rb_root")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{entry.value_():x}\n"
                    for entry in rbtree_preorder_for_each_entry(
                        "struct drgn_test_rb_entry",
                        self.prog["drgn_test_rb_root"].address_of_(),
                        "node",
                    )
                ]
            ),
        )
        self.assertIn("rbtree_preorder_for_each(", cmd.drgn_option.stdout)
        self.assertIn(f"- {offset}", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["entry"].type_.type_name(), "void *")
        self.assertEqual(
            cmd.drgn_option.globals["node"].value_(),
            cmd.drgn_option.globals["entry"].value_() + offset,
        )

    def test_rbtree_entry_type(self):
        cmd = self.check_crash_command("tree -s drgn_test_rb_entry drgn_test_rb_root")
        for node in rbtree_preorder_for_each(
            self.prog["drgn_test_rb_root"].address_of_()
        ):
            self.assertRegex(cmd.stdout, rf"(?m)^{node.value_():x}$")
        self.assertIn("(struct drgn_test_rb_entry){", cmd.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_rb_entry *",
        )

    def test_rbtree_entry_type_and_node_member(self):
        cmd = self.check_crash_command(
            "tree -l -s drgn_test_rb_entry -o drgn_test_rb_entry.node drgn_test_rb_root"
        )
        self.assertEqual(
            re.findall(r"\.value = \(int\)[0-9]+", cmd.stdout),
            [f".value = (int){i}" for i in range(4)],
        )
        self.assertIn("(struct drgn_test_rb_entry){", cmd.stdout)
        self.assertIn("rbtree_inorder_for_each_entry(", cmd.drgn_option.stdout)
        self.assertNotIn("cast(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_rb_entry *",
        )
        self.assertEqual(cmd.drgn_option.globals["entry"].value, 3)

    def test_rbtree_entry_type_and_node_offset(self):
        offset = offsetof(self.prog.type("struct drgn_test_rb_entry"), "node")
        cmd = self.check_crash_command(
            f"tree -l -s drgn_test_rb_entry -o {offset} drgn_test_rb_root"
        )
        self.assertEqual(
            re.findall(r"\.value = \(int\)[0-9]+", cmd.stdout),
            [f".value = (int){i}" for i in range(4)],
        )
        self.assertIn("(struct drgn_test_rb_entry){", cmd.stdout)
        self.assertIn(f"- {offset}", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_rb_entry *",
        )
        self.assertEqual(cmd.drgn_option.globals["entry"].value, 3)

    def test_rbtree_members(self):
        cmd = self.check_crash_command(
            "tree -l -s drgn_test_rb_entry.value -o drgn_test_rb_entry.node drgn_test_rb_root"
        )
        self.assertEqual(
            re.findall(r"\bvalue = \(int\)[0-9]+", cmd.stdout),
            [f"value = (int){i}" for i in range(4)],
        )
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(),
            "struct drgn_test_rb_entry *",
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"].read_(), Object(self.prog, "int", 3)
        )

    def test_rbtree_wrong_entry_type(self):
        cmd = self.check_crash_command(
            "tree -l -s list_head -o drgn_test_rb_entry.node drgn_test_rb_root"
        )
        self.assertIn("(struct list_head){", cmd.stdout)
        self.assertIn("rbtree_inorder_for_each_entry(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["entry"].type_.type_name(), "struct list_head *"
        )

    def test_rbtree_start_address(self):
        address = self.prog["drgn_test_rb_root"].address_
        cmd = self.check_crash_command(f"tree {hex(address)}")
        self.assertEqual(
            cmd.stdout,
            "".join(
                [
                    f"{node.value_():x}\n"
                    for node in rbtree_preorder_for_each(
                        self.prog["drgn_test_rb_root"].address_of_()
                    )
                ]
            ),
        )
        self.assertIn("Object(", cmd.drgn_option.stdout)

    def test_rbtree_root_member(self):
        cmd = self.check_crash_command(
            "tree -l -s drgn_test_rb_entry.value -o drgn_test_rb_entry.node -r drgn_test_rbtree_container_struct.root drgn_test_rbtree_container"
        )
        self.assertEqual(
            re.findall(r"\bvalue = \(int\)[0-9]+", cmd.stdout),
            [f"value = (int){i}" for i in range(2)],
        )
        self.assertIn("root = prog[", cmd.drgn_option.stdout)
        self.assertIn("].root", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_rbtree_container"].root.address_of_(),
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_rbtree_container"].entries[1].value,
        )

    def test_rbtree_root_offset(self):
        offset = offsetof(
            self.prog.type("struct drgn_test_rbtree_container_struct"), "root"
        )
        cmd = self.check_crash_command(
            f"tree -l -s drgn_test_rb_entry.value -o drgn_test_rb_entry.node -r {offset} drgn_test_rbtree_container"
        )
        self.assertEqual(
            re.findall(r"\bvalue = \(int\)[0-9]+", cmd.stdout),
            [f"value = (int){i}" for i in range(2)],
        )
        self.assertIn(f"+ {offset}", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_rbtree_container"].root.address_of_(),
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_rbtree_container"].entries[1].value,
        )

    def test_rbtree_root_member_start_address(self):
        address = self.prog["drgn_test_rbtree_container"].address_
        cmd = self.check_crash_command(
            f"tree -l -s drgn_test_rb_entry.value -o drgn_test_rb_entry.node -r drgn_test_rbtree_container_struct.root {hex(address)}"
        )
        self.assertEqual(
            re.findall(r"\bvalue = \(int\)[0-9]+", cmd.stdout),
            [f"value = (int){i}" for i in range(2)],
        )
        self.assertIn("Object(", cmd.drgn_option.stdout)
        self.assertIn("+= offsetof(", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_rbtree_container"].root.address_of_(),
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_rbtree_container"].entries[1].value,
        )

    def test_rbtree_root_offset_start_address(self):
        address = self.prog["drgn_test_rbtree_container"].address_
        offset = offsetof(
            self.prog.type("struct drgn_test_rbtree_container_struct"), "root"
        )
        cmd = self.check_crash_command(
            f"tree -l -s drgn_test_rb_entry.value -o drgn_test_rb_entry.node -r {offset} {hex(address)}"
        )
        self.assertEqual(
            re.findall(r"\bvalue = \(int\)[0-9]+", cmd.stdout),
            [f"value = (int){i}" for i in range(2)],
        )
        self.assertIn(f"+ {offset}", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_rbtree_container"].root.address_of_(),
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            self.prog["drgn_test_rbtree_container"].entries[1].value,
        )

    def test_radix(self):
        cmd = self.check_crash_command("tree -t radix drgn_test_radix_tree_sparse")
        self.assertEqual(
            cmd.stdout,
            """\
1234
5678
9abc
""",
        )
        self.assertIn("root = prog[", cmd.drgn_option.stdout)
        self.assertIn("radix_tree_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_radix_tree_sparse"].address_of_(),
        )
        self.assertEqual(cmd.drgn_option.globals["index"], 0xFFFFFFFF)
        self.assertIdentical(
            cmd.drgn_option.globals["entry"].read_(),
            Object(self.prog, "void *", 0x9ABC),
        )

    def skip_unless_have_xarray(self):
        if not self.prog["drgn_test_have_xarray"]:
            self.skipTest("kernel does not have XArray")

    def test_xarray(self):
        self.skip_unless_have_xarray()
        cmd = self.check_crash_command("tree -t xarray drgn_test_xarray_sparse")
        self.assertEqual(
            cmd.stdout,
            """\
1234
5678
9abc
""",
        )
        self.assertIn("root = prog[", cmd.drgn_option.stdout)
        self.assertIn("xa_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_xarray_sparse"].address_of_(),
        )
        self.assertEqual(cmd.drgn_option.globals["index"], 0xFFFFFFFF)
        self.assertIdentical(
            cmd.drgn_option.globals["entry"].read_(),
            Object(self.prog, "void *", 0x9ABC),
        )

    def test_xarray_entry_type(self):
        self.skip_unless_have_xarray()
        cmd = self.check_crash_command(
            "tree -t xarray -s drgn_test_xarray_entry drgn_test_xarray_pointers"
        )
        self.assertEqual(cmd.stdout.count("(struct drgn_test_xarray_entry){"), 4)
        self.assertEqual(cmd.drgn_option.globals["index"], 3)
        self.assertIdentical(
            cmd.drgn_option.globals["entry"].read_(),
            self.prog["drgn_test_xarray_entries"][3].address_of_(),
        )

    def test_xarray_members(self):
        self.skip_unless_have_xarray()
        cmd = self.check_crash_command(
            "tree -t xarray -s drgn_test_xarray_entry.value drgn_test_xarray_pointers"
        )
        self.assertEqual(
            re.findall(r"\bvalue = \(int\)[0-9]+", cmd.stdout),
            [f"value = (int){i}" for i in range(4)],
        )
        self.assertIdentical(
            cmd.drgn_option.globals["entry"].read_(),
            self.prog["drgn_test_xarray_entries"][3].address_of_(),
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"].read_(), Object(self.prog, "int", 3)
        )

    def test_maple(self):
        if not self.prog["drgn_test_have_maple_tree"]:
            self.skipTest("kernel does not have maple tree")
        cmd = self.check_crash_command(
            "tree -t maple drgn_test_maple_tree_sparse_ranges"
        )
        self.assertEqual(
            cmd.stdout,
            """\
b0ba000
b0ba001
b0ba002
b0ba003
b0ba004
""",
        )
        self.assertIn("root = prog[", cmd.drgn_option.stdout)
        self.assertIn("mt_for_each(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["root"],
            self.prog["drgn_test_maple_tree_sparse_ranges"].address_of_(),
        )
        self.assertEqual(cmd.drgn_option.globals["first_index"], 81)
        self.assertEqual(cmd.drgn_option.globals["last_index"], 100)
        self.assertIdentical(
            cmd.drgn_option.globals["entry"].read_(),
            Object(self.prog, "void *", 0xB0BA004),
        )

    def test_wrong_type(self):
        cmd = self.run_crash_command_drgn_option(
            "tree drgn_test_radix_tree_sparse", mode="compile"
        )
        self.assertIn("prog.symbol", cmd.stdout)
        self.assertIn("root = Object", cmd.stdout)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import Object
from drgn.helpers.linux.bitmap import bitmap_weight
from drgn.helpers.linux.bitops import for_each_clear_bit, for_each_set_bit, test_bit
from tests import MockProgramTestCase


class TestBitOps(MockProgramTestCase):
    BITMAP = [0xB351BC986648A680, 0x80DDB6615A80BC63]
    # fmt: off
    SET_BITS = [
        7, 9, 10, 13, 15, 19, 22, 25, 26, 29, 30, 35, 36, 39, 42, 43, 44, 45,
        47, 48, 52, 54, 56, 57, 60, 61, 63, 64, 65, 69, 70, 74, 75, 76, 77, 79,
        87, 89, 91, 92, 94, 96, 101, 102, 105, 106, 108, 109, 111, 112, 114,
        115, 116, 118, 119, 127,
    ]
    CLEAR_BITS = [
        0, 1, 2, 3, 4, 5, 6, 8, 11, 12, 14, 16, 17, 18, 20, 21, 23, 24, 27, 28,
        31, 32, 33, 34, 37, 38, 40, 41, 46, 49, 50, 51, 53, 55, 58, 59, 62, 66,
        67, 68, 71, 72, 73, 78, 80, 81, 82, 83, 84, 85, 86, 88, 90, 93, 95, 97,
        98, 99, 100, 103, 104, 107, 110, 113, 117, 120, 121, 122, 123, 124,
        125, 126,
    ]
    # fmt: on

    def test_for_each_set_bit(self):
        bitmap = Object(self.prog, "unsigned long [2]", self.BITMAP)
        self.assertEqual(list(for_each_set_bit(bitmap, 128)), self.SET_BITS)
        self.assertEqual(
            list(for_each_set_bit(bitmap, 101)),
            [bit for bit in self.SET_BITS if bit < 101],
        )

    def test_for_each_clear_bit(self):
        bitmap = Object(self.prog, "unsigned long [2]", self.BITMAP)
        self.assertEqual(list(for_each_clear_bit(bitmap, 128)), self.CLEAR_BITS)
        self.assertEqual(
            list(for_each_clear_bit(bitmap, 100)),
            [bit for bit in self.CLEAR_BITS if bit < 100],
        )

    def test_test_bit(self):
        bitmap = Object(self.prog, "unsigned long [2]", self.BITMAP)
        for bit in self.SET_BITS:
            self.assertTrue(test_bit(bit, bitmap))
        for bit in self.CLEAR_BITS:
            self.assertFalse(test_bit(bit, bitmap))

    def test_bitmap_weight(self):
        bitmap = Object(self.prog, "unsigned long [2]", self.BITMAP)
        self.assertEqual(bitmap_weight(bitmap, 128), len(self.SET_BITS))
        self.assertEqual(
            bitmap_weight(bitmap, 101), sum(1 for bit in self.SET_BITS if bit < 101)
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
from pathlib import Path
import tempfile
import types

from drgn import NULL
from drgn.helpers.linux.ipc import (
    decode_sysv_shm_flags,
    find_sysv_msg_queue,
    find_sysv_sem_array,
    find_sysv_shm,
    for_each_sysv_msg_queue,
    for_each_sysv_sem_array,
    for_each_sysv_shm,
)
from tests.linux_kernel import (
    IPC_PRIVATE,
    IPC_RMID,
    SHM_LOCK,
    LinuxKernelTestCase,
    ftok,
    msgctl,
    msgget,
    semctl,
    semget,
    shmctl,
    shmget,
)

_PROC_SYSVIPC_PATH = Path("/proc/sysvipc")


@contextlib.contextmanager
def tmp_sysv_ipc():
    try:
        msg = []
        sem = []
        shm = []

        with tempfile.NamedTemporaryFile() as tmp_file:
            key = ftok(tmp_file.name, ord("m"))
            msg.append((key, msgget(key)))

            key = ftok(tmp_file.name, ord("e"))
            sem.append((key, semget(key, nsems=4)))

            key = ftok(tmp_file.name, ord("h"))
            shm.append((key, shmget(key, size=64 * 1024)))

        msg.append((IPC_PRIVATE, msgget(IPC_PRIVATE)))
        sem.append((IPC_PRIVATE, semget(IPC_PRIVATE, nsems=4)))
        shm.append((IPC_PRIVATE, shmget(IPC_PRIVATE, size=64 * 1024)))

        shmctl(shm[0][1], SHM_LOCK)

        yield types.SimpleNamespace(msg=msg, sem=sem, shm=shm)
    finally:
        for _, msqid in msg:
            msgctl(msqid, IPC_RMID)
        for _, semid in sem:
            semctl(semid, 0, IPC_RMID)
        for _, shmid in shm:
            shmctl(shmid, IPC_RMID)


def parse_proc_sysvipc_key_and_id(name):
    result = []
    with (_PROC_SYSVIPC_PATH / name).open("r") as f:
        f.readline()  # Skip the header.
        for line in f:
            tokens = line.split()
            result.append((int(tokens[0]), int(tokens[1])))
    assert result
    return result


class TestIpc(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.ipcs = cls.enterClassContext(tmp_sysv_ipc())

    def test_for_each_sysv_msg_queue(self):
        self.assertCountEqual(
            [
                (msg_queue.q_perm.key.value_(), msg_queue.q_perm.id.value_())
                for msg_queue in for_each_sysv_msg_queue(self.prog)
            ],
            parse_proc_sysvipc_key_and_id("msg"),
        )

    def test_find_sysv_msg_queue(self):
        key, id = self.ipcs.msg[0]
        self.assertEqual(find_sysv_msg_queue(self.prog, id).q_perm.key, key)
        self.assertIdentical(
            find_sysv_msg_queue(self.prog, id ^ 0x40000000),
            NULL(self.prog, "struct msg_queue *"),
        )

    def test_for_each_sysv_shm(self):
        self.assertCountEqual(
            [
                (shm.shm_perm.key.value_(), shm.shm_perm.id.value_())
                for shm in for_each_sysv_shm(self.prog)
            ],
            parse_proc_sysvipc_key_and_id("shm"),
        )

    def test_find_sysv_shm(self):
        key, id = self.ipcs.shm[0]
        self.assertEqual(find_sysv_shm(self.prog, id).shm_perm.key, key)
        self.assertIdentical(
            find_sysv_shm(self.prog, id ^ 0x40000000),
            NULL(self.prog, "struct shmid_kernel *"),
        )

    def test_decode_sysv_shm_flags(self):
        id = self.ipcs.shm[0][1]
        self.assertEqual(
            decode_sysv_shm_flags(find_sysv_shm(self.prog, id)), "SHM_LOCKED"
        )

    def test_for_each_sysv_sem_array(self):
        self.assertCountEqual(
            [
                (sem_array.sem_perm.key.value_(), sem_array.sem_perm.id.value_())
                for sem_array in for_each_sysv_sem_array(self.prog)
            ],
            parse_proc_sysvipc_key_and_id("sem"),
        )

    def test_find_sysv_sem_array(self):
        key, id = self.ipcs.sem[0]
        self.assertEqual(find_sysv_sem_array(self.prog, id).sem_perm.key, key)
        self.assertIdentical(
            find_sysv_sem_array(self.prog, id ^ 0x40000000),
            NULL(self.prog, "struct sem_array *"),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import TypeKind, TypeKindSet
from tests import TestCase


class TestTypeKindSet(TestCase):
    def test_repr(self):
        self.assertEqual(
            repr(TypeKindSet()),
            "TypeKindSet()",
        )
        self.assertEqual(
            repr(TypeKindSet({TypeKind.STRUCT})),
            "TypeKindSet({TypeKind.STRUCT})",
        )
        self.assertEqual(
            repr(TypeKindSet({TypeKind.STRUCT, TypeKind.CLASS})),
            "TypeKindSet({TypeKind.STRUCT, TypeKind.CLASS})",
        )

    def test_len(self):
        self.assertEqual(len(TypeKindSet()), 0)
        self.assertEqual(len(TypeKindSet({TypeKind.STRUCT})), 1)
        self.assertEqual(len(TypeKindSet({TypeKind.INT, TypeKind.STRUCT})), 2)

    def test_in(self):
        self.assertNotIn(TypeKind.INT, TypeKindSet())
        self.assertIn(TypeKind.INT, TypeKindSet({TypeKind.INT}))
        self.assertIn(TypeKind.INT, TypeKindSet({TypeKind.INT, TypeKind.STRUCT}))
        self.assertIn(TypeKind.STRUCT, TypeKindSet({TypeKind.INT, TypeKind.STRUCT}))
        self.assertNotIn(TypeKind.UNION, TypeKindSet({TypeKind.INT, TypeKind.STRUCT}))
        self.assertNotIn(0, TypeKindSet({TypeKind.INT}))
        self.assertNotIn(TypeKind.INT.value, TypeKindSet({TypeKind.INT}))

    def test_iter(self):
        for s in (
            set(),
            {TypeKind.VOID},
            {TypeKind.ARRAY, TypeKind.POINTER},
        ):
            self.assertEqual(set(TypeKindSet(s)), s)

    def test_eq(self):
        self.assertTrue(TypeKindSet() == TypeKindSet())
        self.assertFalse(TypeKindSet() == TypeKindSet({TypeKind.BOOL}))
        self.assertTrue(TypeKindSet({TypeKind.FLOAT}) == TypeKindSet({TypeKind.FLOAT}))
        self.assertFalse(
            TypeKindSet({TypeKind.FUNCTION}) == TypeKindSet({TypeKind.BOOL})
        )
        self.assertTrue(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            == TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
        )
        self.assertFalse(
            TypeKindSet({TypeKind.BOOL, TypeKind.FUNCTION})
            == TypeKindSet({TypeKind.BOOL, TypeKind.ARRAY}),
        )
        self.assertTrue(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            == {TypeKind.FLOAT, TypeKind.TYPEDEF}
        )
        self.assertFalse(TypeKindSet() == {"asdf"})
        self.assertFalse(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            == {TypeKind.FLOAT, TypeKind.TYPEDEF, "foo"}
        )
        self.assertFalse(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            == [TypeKind.FLOAT, TypeKind.TYPEDEF]
        )

    def test_ne(self):
        self.assertFalse(TypeKindSet() != TypeKindSet())
        self.assertTrue(TypeKindSet() != TypeKindSet({TypeKind.BOOL}))
        self.assertFalse(TypeKindSet({TypeKind.FLOAT}) != TypeKindSet({TypeKind.FLOAT}))
        self.assertTrue(
            TypeKindSet({TypeKind.FUNCTION}) != TypeKindSet({TypeKind.BOOL})
        )
        self.assertFalse(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            != TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
        )
        self.assertTrue(
            TypeKindSet({TypeKind.BOOL, TypeKind.FUNCTION})
            != TypeKindSet({TypeKind.BOOL, TypeKind.ARRAY}),
        )
        self.assertFalse(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            != {TypeKind.FLOAT, TypeKind.TYPEDEF}
        )
        self.assertTrue(TypeKindSet() != {"asdf"})
        self.assertTrue(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            != {TypeKind.FLOAT, TypeKind.TYPEDEF, "foo"}
        )
        self.assertTrue(
            TypeKindSet({TypeKind.FLOAT, TypeKind.TYPEDEF})
            != [TypeKind.FLOAT, TypeKind.TYPEDEF]
        )

    def test_lt(self):
        self.assertFalse(TypeKindSet() < TypeKindSet())
        self.assertTrue(TypeKindSet() < TypeKindSet({TypeKind.BOOL}))
        self.assertFalse(TypeKindSet({TypeKind.BOOL}) < TypeKindSet())
        self.assertFalse(TypeKindSet({TypeKind.BOOL}) < TypeKindSet({TypeKind.BOOL}))
        self.assertTrue(
            TypeKindSet({TypeKind.BOOL}) < TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT})
        )
        self.assertFalse(
            TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT}) < TypeKindSet({TypeKind.BOOL})
        )
        self.assertFalse(TypeKindSet({TypeKind.INT}) < TypeKindSet({TypeKind.BOOL}))
        self.assertTrue(
            TypeKindSet({TypeKind.BOOL}) < {TypeKind.BOOL, TypeKind.FLOAT, "foo"}
        )
        with self.assertRaises(TypeError):
            TypeKindSet({TypeKind.BOOL}) < [TypeKind.BOOL, TypeKind.FLOAT]

    def test_gt(self):
        self.assertFalse(TypeKindSet() > TypeKindSet())
        self.assertFalse(TypeKindSet() > TypeKindSet({TypeKind.BOOL}))
        self.assertTrue(TypeKindSet({TypeKind.BOOL}) > TypeKindSet())
        self.assertFalse(TypeKindSet({TypeKind.BOOL}) > TypeKindSet({TypeKind.BOOL}))
        self.assertFalse(
            TypeKindSet({TypeKind.BOOL}) > TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT})
        )
        self.assertTrue(
            TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT}) > TypeKindSet({TypeKind.BOOL})
        )
        self.assertFalse(TypeKindSet({TypeKind.INT}) > TypeKindSet({TypeKind.BOOL}))
        self.assertFalse(
            TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT}) > {TypeKind.BOOL, "foo"}
        )
        with self.assertRaises(TypeError):
            TypeKindSet({TypeKind.BOOL}) > []

    def test_le(self):
        self.assertTrue(TypeKindSet() <= TypeKindSet())
        self.assertTrue(TypeKindSet() <= TypeKindSet({TypeKind.BOOL}))
        self.assertFalse(TypeKindSet({TypeKind.BOOL}) <= TypeKindSet())
        self.assertTrue(TypeKindSet({TypeKind.BOOL}) <= TypeKindSet({TypeKind.BOOL}))
        self.assertTrue(
            TypeKindSet({TypeKind.BOOL}) <= TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT})
        )
        self.assertFalse(
            TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT}) <= TypeKindSet({TypeKind.BOOL})
        )
        self.assertFalse(TypeKindSet({TypeKind.INT}) <= TypeKindSet({TypeKind.BOOL}))
        self.assertTrue(TypeKindSet({TypeKind.BOOL}) <= {TypeKind.BOOL, "foo"})
        with self.assertRaises(TypeError):
            TypeKindSet({TypeKind.BOOL}) <= [TypeKind.BOOL, TypeKind.FLOAT]

    def test_ge(self):
        self.assertTrue(TypeKindSet() >= TypeKindSet())
        self.assertFalse(TypeKindSet() >= TypeKindSet({TypeKind.BOOL}))
        self.assertTrue(TypeKindSet({TypeKind.BOOL}) >= TypeKindSet())
        self.assertTrue(TypeKindSet({TypeKind.BOOL}) >= TypeKindSet({TypeKind.BOOL}))
        self.assertFalse(
            TypeKindSet({TypeKind.BOOL}) >= TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT})
        )
        self.assertTrue(
            TypeKindSet({TypeKind.BOOL, TypeKind.FLOAT}) >= TypeKindSet({TypeKind.BOOL})
        )
        self.assertFalse(TypeKindSet({TypeKind.INT}) >= TypeKindSet({TypeKind.BOOL}))
        self.assertFalse(TypeKindSet({TypeKind.BOOL}) >= {TypeKind.BOOL, "foo"})
        with self.assertRaises(TypeError):
            TypeKindSet({TypeKind.BOOL}) >= [TypeKind.BOOL]

    def test_isdisjoint(self):
        self.assertTrue(TypeKindSet().isdisjoint(TypeKindSet()))
        self.assertTrue(
            TypeKindSet({TypeKind.POINTER}).isdisjoint(TypeKindSet({TypeKind.ARRAY}))
        )
        self.assertFalse(
            TypeKindSet({TypeKind.POINTER}).isdisjoint(TypeKindSet({TypeKind.POINTER}))
        )
        self.assertFalse(
            TypeKindSet({TypeKind.ARRAY}).isdisjoint(
                TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY})
            )
        )
        self.assertFalse(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}).isdisjoint(
                TypeKindSet({TypeKind.ARRAY})
            )
        )
        self.assertTrue(TypeKindSet({TypeKind.POINTER}).isdisjoint({TypeKind.ARRAY}))
        self.assertTrue(TypeKindSet({TypeKind.POINTER}).isdisjoint([TypeKind.ARRAY]))
        self.assertFalse(TypeKindSet({TypeKind.POINTER}).isdisjoint({TypeKind.POINTER}))
        self.assertFalse(TypeKindSet({TypeKind.POINTER}).isdisjoint([TypeKind.POINTER]))

    def test_sub(self):
        self.assertEqual(TypeKindSet() - TypeKindSet(), TypeKindSet())
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) - TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) - TypeKindSet({TypeKind.POINTER}),
            TypeKindSet(),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.ARRAY})
            - TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
            TypeKindSet(),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY})
            - TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}) - {TypeKind.ARRAY},
            TypeKindSet({TypeKind.POINTER}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}) - {TypeKind.ARRAY, "asdf"},
            TypeKindSet({TypeKind.POINTER}),
        )

    def test_and(self):
        self.assertEqual(TypeKindSet() & TypeKindSet(), TypeKindSet())
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) & TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet(),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) & TypeKindSet({TypeKind.POINTER}),
            TypeKindSet({TypeKind.POINTER}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.ARRAY})
            & TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
            TypeKindSet({TypeKind.ARRAY}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY})
            & TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet({TypeKind.ARRAY}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}) & {TypeKind.ARRAY},
            TypeKindSet({TypeKind.ARRAY}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY})
            & {TypeKind.POINTER, "asdf"},
            TypeKindSet({TypeKind.POINTER}),
        )

    def test_xor(self):
        self.assertEqual(TypeKindSet() ^ TypeKindSet(), TypeKindSet())
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) ^ TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) ^ TypeKindSet({TypeKind.POINTER}),
            TypeKindSet(),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.ARRAY})
            ^ TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY})
            ^ TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}) ^ {TypeKind.ARRAY},
            TypeKindSet({TypeKind.POINTER}),
        )
        with self.assertRaises(TypeError):
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}) ^ {TypeKind.ARRAY, "asdf"}

    def test_or(self):
        self.assertEqual(TypeKindSet() | TypeKindSet(), TypeKindSet())
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) | TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) | TypeKindSet({TypeKind.POINTER}),
            TypeKindSet({TypeKind.POINTER}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.ARRAY})
            | TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY})
            | TypeKindSet({TypeKind.ARRAY}),
            TypeKindSet({TypeKind.POINTER, TypeKind.ARRAY}),
        )
        self.assertEqual(
            TypeKindSet({TypeKind.POINTER}) | {TypeKind.ARRAY},
            {TypeKind.POINTER, TypeKind.ARRAY},
        )
        with self.assertRaises(TypeError):
            TypeKindSet({TypeKind.POINTER}) | {TypeKind.ARRAY, "asdf"}

    def test_all_kinds(self):
        self.assertEqual(set(TypeKindSet(TypeKind)), set(TypeKind))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


import argparse
import keyword
import re
import sys
from typing import Dict, List, NamedTuple, Sequence, TextIO, cast


class DwarfConstant(NamedTuple):
    name: str
    value: int


class DwarfConstantType(NamedTuple):
    name: str
    constants: Sequence[DwarfConstant]


def parse_dwarf_constants(f: TextIO) -> Sequence[DwarfConstantType]:
    types: Dict[str, Dict[str, int]] = {
        type_name: {}
        for type_name in (
            "DW_ACCESS",
            "DW_ADDR",
            "DW_AT",
            "DW_ATE",
            "DW_CC",
            "DW_CFA",
            "DW_CHILDREN",
            "DW_DEFAULTED",
            "DW_DS",
            "DW_DSC",
            "DW_EH_PE",
            "DW_END",
            "DW_FORM",
            "DW_ID",
            "DW_IDX",
            "DW_INL",
            "DW_LANG",
            "DW_LLE",
            "DW_LNCT",
            "DW_LNE",
            "DW_LNS",
            "DW_MACINFO",
            "DW_MACRO",
            "DW_OP",
            "DW_ORD",
            "DW_RLE",
            "DW_SECT",
            "DW_TAG",
            "DW_UT",
            "DW_VIRTUALITY",
            "DW_VIS",
        )
    }

    for match in re.finditer(
        r"^\s*#\s*define\s+(" + "|".join(types) + r")_(\w+)\s+(\S+)",
        sys.stdin.read(),
        flags=re.MULTILINE,
    ):
        type_name = match.group(1)
        name = match.group(2)
        value = int(match.group(3), 0)
        if (type_name, name) in {
            # Typos in the wild that libdwarf includes but we don't want.
            ("DW_AT", "stride"),  # "DWARF3 (do not use)"
            ("DW_CFA", "low_user"),  # "Incorrect spelling, do not use"
            ("DW_TAG", "namelist_items"),  # "SGI misspelling/typo"
            ("DW_TAG", "template_type_param"),  # "DWARF2 inconsistent"
            ("DW_TAG", "template_value_param"),  # "DWARF2 inconsistent"
            # libdwarf probably included this one to be consistent with the
            # standard DWARF template_foo_parameter names, but it's called
            # DW_TAG_GNU_template_template_param everywhere else.
            ("DW_TAG", "GNU_template_template_parameter"),
            # This name isn't mentioned in any version of the DWARF standard.
            ("DW_CFA", "extended"),
        }:
            continue
        # Typos in libdwarf itself.
        elif (type_name, name) == ("DW_CFA", "high_user"):
            name = "hi_user"
        elif (type_name, name) == ("DW_IDX", "hi_user"):
            value = 0x3FFF
        elif (type_name, name) == ("DW_LANG", "Haskel"):
            name = "Haskell"
        if types[type_name].setdefault(name, value) != value:
            raise ValueError(f"{type_name}_{name} redefined with different value")

    result = [
        DwarfConstantType(
            name=type_name,
            constants=[DwarfConstant(name, value) for name, value in constants.items()],
        )
        for type_name, constants in types.items()
    ]

    def insert_after(
        type_name: str, after_name: str, insert_constant: DwarfConstant
    ) -> None:
        for constant_type in result:
            if constant_type.name == type_name:
                break
        else:
            raise ValueError()
        constants = cast(List[DwarfConstant], constant_type.constants)
        for i, constant in enumerate(constants):
            if constant.name == after_name:
                break
        else:
            raise ValueError()
        constants.insert(i + 1, insert_constant)

    insert_after("DW_EH_PE", "sdata8", DwarfConstant("signed", 0x8))
    insert_after("DW_EH_PE", "aligned", DwarfConstant("indirect", 0x80))

    return result


_DWARF_CONSTANTS_WANT_STR = {"DW_OP", "DW_TAG"}


def gen_dwarf_constants_h(
    dwarf_constants: Sequence[DwarfConstantType], f: TextIO
) -> None:
    f.write(
        """\
// Copyright (c) Meta Platforms, Inc. and affiliates.
// SPDX-License-Identifier: LGPL-2.1-or-later
// Generated by scripts/gen_dwarf_constants.py.

/**
 * @file
 *
 * DWARF constant definitions.
 *
 * This file defines the following for each known DWARF constant type:
 *
 * 1. An X macro defining all of the known names and values of the type:
 *    `DW_FOO_DEFINITIONS`.
 * 2. Enumerators defining the constants: `DW_FOO_a`, `DW_FOO_b`, etc.
 * 3. For select types, a function to translate a value to its name:
 *   `dw_foo_str()`.
 */

#ifndef DWARF_CONSTANTS_H
#define DWARF_CONSTANTS_H

#define X(name, value) name = value,
"""
    )
    for constant_type in dwarf_constants:
        f.write(
            f"""
#define {constant_type.name}_DEFINITIONS \\
"""
        )
        for i, constant in enumerate(constant_type.constants):
            end = " \\" if i < len(constant_type.constants) - 1 else ""
            f.write(
                f"\tX({constant_type.name}_{constant.name}, 0x{constant.value:x}){end}\n"
            )
        f.write(f"enum {{ {constant_type.name}_DEFINITIONS }};\n")
        if constant_type.name in _DWARF_CONSTANTS_WANT_STR:
            f.write(
                f"""\
#define {constant_type.name}_STR_UNKNOWN_FORMAT "{constant_type.name}_<0x%x>"
#define {constant_type.name}_STR_BUF_LEN (sizeof({constant_type.name}_STR_UNKNOWN_FORMAT) - 2 + 2 * sizeof(int))
/**
 * Get the name of a `{constant_type.name}` value.
 *
 * @return Static string if the value is known or @p buf if the value is
 * unknown.
 */
const char *{constant_type.name.lower()}_str(int value, char buf[static {constant_type.name}_STR_BUF_LEN]);
"""
            )

    f.write(
        """
#undef X

#endif /* DWARF_CONSTANTS_H */
"""
    )


def gen_dwarf_constants_c(
    dwarf_constants: Sequence[DwarfConstantType], f: TextIO
) -> None:
    f.write(
        """\
// Copyright (c) Meta Platforms, Inc. and affiliates.
// SPDX-License-Identifier: LGPL-2.1-or-later
// Generated by scripts/gen_dwarf_constants.py.

#include <stdio.h>

#include "dwarf_constants.h"

#define X(name, _) if (value == name) return #name;
"""
    )
    for constant_type in dwarf_constants:
        if constant_type.name in _DWARF_CONSTANTS_WANT_STR:
            f.write(
                f"""
const char *{constant_type.name.lower()}_str(int value, char buf[static {constant_type.name}_STR_BUF_LEN])
{{
	{constant_type.name}_DEFINITIONS
	snprintf(buf, {constant_type.name}_STR_BUF_LEN, {constant_type.name}_STR_UNKNOWN_FORMAT, value);
	return buf;
}}
"""
            )

    f.write(
        """
#undef X
"""
    )


def gen_tests_dwarf_py(dwarf_constants: Sequence[DwarfConstantType], f: TextIO) -> None:
    f.write(
        """\
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
# Generated by scripts/gen_dwarf_constants.py.

import enum
"""
    )
    for constant_type in dwarf_constants:
        f.write(f"\n\nclass {constant_type.name}(enum.IntEnum):\n")
        for constant in constant_type.constants:
            name = constant.name
            if keyword.iskeyword(name):
                name += "_"
            f.write(f"    {name} = 0x{constant.value:X}")
            if name == "name":
                f.write("  # type: ignore")
            f.write("\n")


def main() -> None:
    argparse.ArgumentParser(
        description="Generate libdrgn/dwarf_constants.h, libdrgn/dwarf_constants.c, and tests/dwarf.py from libdwarf/src/lib/libdwarf/dwarf.h (read from standard input)"
    ).parse_args()

    dwarf_constants = parse_dwarf_constants(sys.stdin)
    with open("libdrgn/dwarf_constants.h", "w") as f:
        gen_dwarf_constants_h(dwarf_constants, f)
    with open("libdrgn/dwarf_constants.c", "w") as f:
        gen_dwarf_constants_c(dwarf_constants, f)
    with open("tests/dwarf.py", "w") as f:
        gen_tests_dwarf_py(dwarf_constants, f)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import array
import contextlib
import ctypes
import errno
import io
import mmap
import os
from pathlib import Path
import re
import socket
import subprocess
import sys
import tempfile
import unittest

from drgn import container_of
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import (
    CLONE_NEWNS,
    CLONE_NEWUSER,
    MS_NODEV,
    MS_NOEXEC,
    MS_NOSUID,
    LinuxKernelTestCase,
    fallocate,
    fork_and_stop,
    iter_mounts,
    losetup,
    mkswap,
    mount,
    perf_event_attr,
    perf_event_open,
    skip_unless_have_test_disk,
    swapoff,
    swapon,
    umount,
    unshare,
)
from tools.fsrefs import main

UPROBE_TYPE_PATH = Path("/sys/bus/event_source/devices/uprobe/type")


class TestFsRefs(LinuxKernelTestCase):
    def setUp(self):
        super().setUp()
        self._tmp = Path(self.enterContext(tempfile.TemporaryDirectory()))

    def run_and_capture(self, *args):
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            main(self.prog, args)
        return f.getvalue()

    def test_fd(self):
        path = self._tmp / "file"
        fd = os.open(path, os.O_CREAT | os.O_WRONLY, 0o600)
        try:
            self.assertRegex(
                self.run_and_capture("--check", "tasks", "--inode", str(path)),
                rf"pid {os.getpid()} \(.*\) fd {fd} ",
            )
        finally:
            os.close(fd)

    def test_dereference(self):
        file = self._tmp / "file"
        link = self._tmp / "link"
        file.touch()
        link.symlink_to("file")

        file_fd = os.open(file, os.O_CREAT | os.O_WRONLY, 0o600)
        try:
            link_fd = os.open(link, os.O_PATH | os.O_NOFOLLOW)
            try:
                output = self.run_and_capture("--check", "tasks", "--inode", str(link))
                self.assertNotRegex(
                    output,
                    rf"pid {os.getpid()} \(.*\) fd {file_fd} ",
                )
                self.assertRegex(
                    output,
                    rf"pid {os.getpid()} \(.*\) fd {link_fd} ",
                )

                output = self.run_and_capture(
                    "--check", "tasks", "--inode", str(link), "--dereference"
                )
                self.assertRegex(
                    output,
                    rf"pid {os.getpid()} \(.*\) fd {file_fd} ",
                )
                self.assertNotRegex(
                    output,
                    rf"pid {os.getpid()} \(.*\) fd {link_fd} ",
                )
            finally:
                os.close(link_fd)
        finally:
            os.close(file_fd)

    def test_cwd(self):
        def mkdir_and_chdir(path, mode):
            os.mkdir(path, mode)
            os.chdir(path)

        path = self._tmp / "dir"
        with fork_and_stop(mkdir_and_chdir, path, 0o600) as (pid, _):
            self.assertRegex(
                self.run_and_capture("--check", "tasks", "--inode", str(path)),
                rf"pid {pid} \(.*\) cwd ",
            )

    def test_root(self):
        def mkdir_and_chroot(path, mode):
            os.mkdir(path, mode)
            os.chroot(path)

        path = self._tmp / "dir"
        with fork_and_stop(mkdir_and_chroot, path, 0o600) as (pid, _):
            self.assertRegex(
                self.run_and_capture("--check", "tasks", "--inode", str(path)),
                rf"pid {pid} \(.*\) root ",
            )

    def test_exe(self):
        self.assertRegex(
            self.run_and_capture(
                "--check", "tasks", "--inode", sys.executable, "--dereference"
            ),
            rf"pid {os.getpid()} \(.*\) exe ",
        )

    def test_vma(self):
        path = self._tmp / "file"
        with open(path, "w+b") as f:
            os.ftruncate(f.fileno(), mmap.PAGESIZE)
            # Note: this dups the file descriptor internally.
            with mmap.mmap(f.fileno(), mmap.PAGESIZE) as map:
                f.close()
                start = ctypes.addressof(ctypes.c_char.from_buffer(map))
                end = start + mmap.PAGESIZE
                self.assertRegex(
                    self.run_and_capture("--check", "tasks", "--inode", str(path)),
                    rf"pid {os.getpid()} \(.*\) vma {hex(start)}-{hex(end)} ",
                )

    def test_inode_pointer(self):
        self.assertRegex(
            self.run_and_capture(
                "--check",
                "tasks",
                "--inode-pointer",
                hex(find_task(self.prog, os.getpid()).mm.exe_file.f_inode),
                "--dereference",
            ),
            rf"pid {os.getpid()} \(.*\) exe ",
        )

    def test_super_block(self):
        with contextlib.ExitStack() as exit_stack:
            mount("tmpfs", self._tmp, "tmpfs")
            exit_stack.callback(umount, self._tmp)

            pid, _ = exit_stack.enter_context(fork_and_stop(unshare, CLONE_NEWNS))

            path1 = self._tmp / "file1"
            fd1 = os.open(path1, os.O_CREAT | os.O_WRONLY, 0o600)
            exit_stack.callback(os.close, fd1)

            path2 = self._tmp / "file2"
            fd2 = os.open(path2, os.O_CREAT | os.O_WRONLY, 0o600)
            exit_stack.callback(os.close, fd2)

            output = self.run_and_capture(
                "--check", "mounts", "--check", "tasks", "--super-block", str(self._tmp)
            )

            with self.subTest("mount"):
                self.assertIn(f"mount {self._tmp} (struct mount", output)

            with self.subTest("mount in namespace"):
                ino = Path(f"/proc/{pid}/ns/mnt").stat().st_ino
                self.assertIn(f"mount {self._tmp} (mount namespace {ino}) ", output)

            with self.subTest("fd"):
                self.assertRegex(
                    output,
                    rf"pid {os.getpid()} \(.*\) fd {fd1} \(struct file \*\)0x[0-9a-f]+ {re.escape(str(path1))}",
                )
                self.assertRegex(
                    output,
                    rf"pid {os.getpid()} \(.*\) fd {fd2} \(struct file \*\)0x[0-9a-f]+ {re.escape(str(path2))}",
                )

            with self.subTest("super_block_pointer"):
                self.assertIn(
                    f"mount {self._tmp} ",
                    self.run_and_capture(
                        "--check",
                        "mounts",
                        "--super-block-pointer",
                        hex(fget(find_task(self.prog, os.getpid()), fd1).f_inode.i_sb),
                    ),
                )

    @skip_unless_have_test_disk
    def test_super_block_on_block_device(self):
        disk = os.environ["DRGN_TEST_DISK"]
        for fstype, mkfs in (
            ("ext2", ("mke2fs", "-qF")),
            ("btrfs", ("mkfs.btrfs", "-qf", "-s", str(mmap.PAGESIZE))),
        ):
            with self.subTest(fstype=fstype):
                subprocess.check_call([*mkfs, disk])

                with contextlib.ExitStack() as exit_stack:
                    mount(disk, self._tmp, fstype)
                    exit_stack.callback(umount, self._tmp)

                    self.assertIn(
                        f"mount {self._tmp} ",
                        self.run_and_capture(
                            "--check",
                            "mounts",
                            "--dereference",
                            "--super-block-on-block-device",
                            str(disk),
                        ),
                    )

    @skip_unless_have_test_disk
    def test_btrfs_subvolume(self):
        disk = os.environ["DRGN_TEST_DISK"]
        with contextlib.ExitStack() as exit_stack:
            subprocess.check_call(["mkfs.btrfs", "-qf", "-s", str(mmap.PAGESIZE), disk])

            mount(disk, self._tmp, "btrfs")
            exit_stack.callback(umount, self._tmp)

            subvol = self._tmp / "subvol"
            subprocess.check_call(["btrfs", "-q", "subvolume", "create", subvol])

            top_file = self._tmp / "file"
            top_fd = os.open(top_file, os.O_CREAT | os.O_WRONLY, 0o600)
            exit_stack.callback(os.close, top_fd)
            top_regex = rf"pid {os.getpid()} \(.*\) fd {top_fd} \(struct file \*\)0x[0-9a-f]+ {re.escape(str(top_file))}"

            subvol_file = subvol / "file"
            subvol_fd = os.open(subvol_file, os.O_CREAT | os.O_WRONLY, 0o600)
            exit_stack.callback(os.close, subvol_fd)
            subvol_regex = rf"pid {os.getpid()} \(.*\) fd {subvol_fd} \(struct file \*\)0x[0-9a-f]+ {re.escape(str(subvol_file))}"

            subvol_output = self.run_and_capture(
                "--check", "tasks", "--btrfs-subvolume", str(subvol)
            )
            self.assertRegex(subvol_output, subvol_regex)
            self.assertNotRegex(subvol_output, top_regex)

            top_root = container_of(
                fget(find_task(self.prog, os.getpid()), top_fd).f_inode,
                "struct btrfs_inode",
                "vfs_inode",
            ).root
            top_output = self.run_and_capture(
                "--check", "tasks", "--btrfs-subvolume-pointer", hex(top_root)
            )
            self.assertRegex(top_output, top_regex)
            self.assertNotRegex(top_output, subvol_regex)

    def test_not_btrfs(self):
        with self.assertRaises(SystemExit) as cm:
            main(self.prog, ["--check", "tasks", "--btrfs-subvolume", "/proc"])
        self.assertIn("not on Btrfs", cm.exception.code)

    def test_binfmt_misc(self):
        for mnt in iter_mounts():
            if mnt.fstype == "binfmt_misc":
                break
        else:
            self.skipTest("binfmt_misc not mounted")

        path = self._tmp / "file"
        path.touch()
        path.chmod(0o700)
        try:
            id = os.urandom(20)
            name = f"drgntest_{id.hex()}"
            encoded_id = "".join([f"\\x{byte:02x}" for byte in id])
            (mnt.mount_point / "register").write_text(
                f":{name}:M::{encoded_id}::{path}:F"
            )

            self.assertIn(
                f"binfmt_misc {name} ",
                self.run_and_capture("--check", "binfmt_misc", "--inode", str(path)),
            )
        finally:
            try:
                with open(mnt.mount_point / name, "r+") as f:
                    f.write("-1")
            except FileNotFoundError:
                pass

    def test_binfmt_misc_in_user_ns(self):
        id = os.urandom(20)
        name = f"drgntest_{id.hex()}"

        def setup_binfmt_misc_in_userns(path):
            try:
                unshare(CLONE_NEWUSER | CLONE_NEWNS)
            except OSError as e:
                if e.errno == errno.EINVAL:
                    return "kernel does not support user namespaces (CONFIG_USER_NS)"
                else:
                    raise
            Path("/proc/self/uid_map").write_text("0 0 1")
            Path("/proc/self/setgroups").write_text("deny")
            Path("/proc/self/gid_map").write_text("0 0 1")

            mount_point = path.parent / "binfmt_misc"
            mount_point.mkdir()
            try:
                mount(
                    "binfmt_misc",
                    mount_point,
                    "binfmt_misc",
                    MS_NOSUID | MS_NODEV | MS_NOEXEC,
                )
            except OSError as e:
                if e.errno == errno.ENODEV:
                    return "kernel does not support binfmt_misc (CONFIG_BINFMT_MISC)"
                elif e.errno == errno.EPERM:
                    return "kernel does not support sandboxed binfmt_misc mounts"
                else:
                    raise

            path.touch()
            path.chmod(0o700)
            encoded_id = "".join([f"\\x{byte:02x}" for byte in id])
            (mount_point / "register").write_text(f":{name}:M::{encoded_id}::{path}:F")

        path = self._tmp / "file"
        with fork_and_stop(setup_binfmt_misc_in_userns, path) as (pid, skip):
            if skip:
                self.skipTest(skip)
            ino = Path(f"/proc/{pid}/ns/user").stat().st_ino
            self.assertIn(
                f"binfmt_misc (user namespace {ino}) {name} ",
                self.run_and_capture("--check", "binfmt_misc", "--inode", str(path)),
            )

    def test_loop_device(self):
        path = self._tmp / "file"
        with open(path, "wb") as f:
            os.ftruncate(f.fileno(), 1024 * 1024)
            with losetup(f.fileno()) as loop_file:
                f.close()
                number = int(loop_file.name.replace("/dev/loop", ""))
                self.assertIn(
                    f"loop device {number} ",
                    self.run_and_capture("--check", "loop", "--inode", str(path)),
                )

    @skip_unless_have_test_disk
    def test_swap_file(self):
        disk = os.environ["DRGN_TEST_DISK"]
        subprocess.check_call(["mke2fs", "-qF", disk])

        with contextlib.ExitStack() as exit_stack:
            mount(disk, self._tmp, "ext2")
            exit_stack.callback(umount, self._tmp)

            path = self._tmp / "swap_file"
            fallocate(path, 0, 1024 * 1024)
            mkswap(path)
            swapon(path)
            exit_stack.callback(swapoff, path)

            self.assertIn(
                "swap file (struct swap_info_struct *)",
                self.run_and_capture("--check", "swap", "--inode", str(path)),
            )

    def test_uprobe_event(self):
        for mnt in iter_mounts():
            if mnt.fstype == "tracefs":
                break
        else:
            self.skipTest("tracefs not mounted")
        uprobe_events = mnt.mount_point / "uprobe_events"
        if not uprobe_events.exists():
            self.skipTest(
                "kernel does not support uprobe events (CONFIG_UPROBE_EVENTS)"
            )

        def uprobe_events_append(s):
            # open(..., "a") tries lseek(..., SEEK_END), which fails with
            # EINVAL.
            with open(os.open(uprobe_events, os.O_WRONLY | os.O_APPEND), "w") as f:
                f.write(s)

        path = self._tmp / "file"
        path.touch()

        probe_name = f"drgntest_{os.urandom(20).hex()}"
        retprobe_name = f"drgntest_{os.urandom(20).hex()}"
        with contextlib.ExitStack() as exit_stack:
            uprobe_events_append(f"p:{probe_name} {path}:0\n")
            exit_stack.callback(uprobe_events_append, f"-:{probe_name}\n")
            uprobe_events_append(f"r:{retprobe_name} {path}:0\n")
            exit_stack.callback(uprobe_events_append, f"-:{retprobe_name}\n")

            instance = Path(tempfile.mkdtemp(dir=mnt.mount_point / "instances"))
            exit_stack.callback(instance.rmdir)

            (instance / "events/uprobes" / probe_name / "enable").write_text("1")
            (instance / "events/uprobes" / retprobe_name / "enable").write_text("1")

            output = self.run_and_capture("--check", "uprobes", "--inode", str(path))
            self.assertIn(f"uprobe event p:uprobes/{probe_name} ", output)
            self.assertIn(f"uprobe event r:uprobes/{retprobe_name} ", output)

    @unittest.skipUnless(
        UPROBE_TYPE_PATH.exists(), "kernel does not support perf_uprobe"
    )
    def test_perf_uprobe(self):
        path = self._tmp / "file"
        path.touch()

        attr = perf_event_attr()
        attr.type = int(UPROBE_TYPE_PATH.read_text())
        ctypes_path = ctypes.c_char_p(os.fsencode(path))
        attr.uprobe_path = ctypes.cast(ctypes_path, ctypes.c_void_p).value
        fd = perf_event_open(attr, -1, min(os.sched_getaffinity(0)))
        try:
            self.assertIn(
                f"perf uprobe (owned by pid {os.getpid()}",
                self.run_and_capture("--check", "uprobes", "--inode", str(path)),
            )
        finally:
            os.close(fd)

    @unittest.skipUnless(
        UPROBE_TYPE_PATH.exists(), "kernel does not support perf_uprobe"
    )
    def test_perf_uprobe_no_owner(self):
        path = self._tmp / "file"
        path.touch()

        sock1, sock2 = socket.socketpair()
        try:
            # Create a perf event in a process, send it over a Unix socket to
            # keep it alive, then die.
            pid = os.fork()
            if pid == 0:
                try:
                    attr = perf_event_attr()
                    attr.type = int(UPROBE_TYPE_PATH.read_text())
                    ctypes_path = ctypes.c_char_p(os.fsencode(path))
                    attr.uprobe_path = ctypes.cast(ctypes_path, ctypes.c_void_p).value
                    fd = perf_event_open(attr, -1, min(os.sched_getaffinity(0)))
                    sock2.sendmsg(
                        [b"\0"],
                        [
                            (
                                socket.SOL_SOCKET,
                                socket.SCM_RIGHTS,
                                array.array("i", [fd]),
                            )
                        ],
                    )
                finally:
                    os._exit(0)

            os.waitpid(pid, 0)
            self.assertIn(
                "perf uprobe (no owner)",
                self.run_and_capture("--check", "uprobes", "--inode", str(path)),
            )
        finally:
            sock1.close()
            sock2.close()
